{$A+,B-,C+,D+,E-,F-,G+,H+,I+,J-,K-,L+,M-,N+,O-,P+,Q-,R+,S-,T-,U-,V+,W-,X+,Y+,Z1}
//{ $ MINSTACKSIZE $00004000}
//{ $ MAXSTACKSIZE $00100000}
//{ $ IMAGEBASE $00400000}
//{ $ APPTYPE GUI}

// (c) 2003, Ivan Yanikov, dobrokot 'at' gmail.com 

unit frmMain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls,
  jpeg, mat, OpenGL, Contnrs;

procedure glBindTexture(target: GLenum; texture: GLuint);stdcall;external 'opengl32.dll';
//Mistake in original OpengGL.pas in 'gluBuild2DMipmaps' defenition:
function gluBuild2DMipmaps(Target: GLenum; Components, Width, Height: GLint; Format, atype: GLenum; Data: Pointer): GLint; stdcall; external 'glu32.DLL';
procedure glGenTextures(n: integer; var textures: GLuint);stdcall;external 'opengl32.dll';


type
  TfrmGL= class;

  TMesh= class
    procedure Draw; virtual; abstract;
  end;

  TLight=class
    pos: Vector;
    col: Vector; //r,g,b
    procedure LoadToGL(LIGHTx: integer);
  end;

  TObj=class
    frmGL: TfrmGL;
    SubObj: TObjectList;
    constructor Create(frmGL: TfrmGL);
    destructor Destroy; override;
    procedure Draw; virtual;
  end;

  TShCaster=class(TObj)
    //procedure Draw; override; abstract;
    procedure CastShadows(const l: Vector); virtual; abstract;
  end;

  TSphere=class (TShCaster)
    r:RealType;
    pos:Vector;
    list: GLuint;
    Tex: GLuint;
    constructor Create(frmGL: TfrmGL; sr: RealType; TexName: string);
    procedure Draw; override;
    procedure CastShadows(const l: Vector); override;
  end;

  TPseudoEarth=class (TObj)
    Orbit: RealType;
    Year: RealType;
    procedure Draw; override;
    constructor Create(frmGL: TfrmGL);
  end;

  TEarth=class(TSphere)
    procedure Draw; override;
  end;

  TAtmosphere=class(TSphere)
    procedure Draw; override;
  end;

  TMoon=class(TSphere)
    Orbit: RealType;
    Year: RealType;
    constructor Create(frmGL: TfrmGL; sr:RealType; TexName: string);
    procedure Draw; override;
  end;


  TSky=class(TSphere)
    procedure Draw; override;
  end;

  Particle=object
    size: RealType;
    pos: Vector;
    v: Vector;
    max, live: RealType;
    //MODELVIEW in gl should be identity
    //m should be real MODELVIEW.
    //OpenGL should be in glBegin(GL_QUADS) mode
    procedure Draw(var m:Matrix);
    procedure Next(dt: RealType);
    procedure Born;
  end;

  TSun=class(TObj)
    ps: array of Particle;
    n: integer;
    tex_id: GLuint;
    constructor Create(frmGL: TfrmGL);
    destructor Destroy; override;
    procedure Draw; override;
  end;


  TfrmGL = class(TForm)
    procedure FormCreate(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
  public
    Scene: TObj;
      PseudoEarth: TPseudoEarth;
        Earth: TEarth;
        Moon: TMoon;
        Atmosphere: TAtmosphere;
      Sun: TSun;
      Sky: TSky;
    Light: TLight;
    LightPos: Vector;
    Time, LastTime: double;
    CurrMatrix: Matrix;
    Distance: RealType;
    procedure Save(var SaveTo:Matrix);
    procedure Restore(var RestoreFrom:Matrix);
    procedure Mul(const m:Matrix);
  private
    TimeToAwake: RealType;
    sp: integer;
    DC: HDC;
    hrc: HGLRC;
    StartTime: cardinal;
    MouseMatrix: Matrix;
    OldMouseY, OldMouseX:integer;
    LeftMouse, RightMouse: boolean;
    ShadowCasters: TObjectList;
    procedure GenShadows;
    procedure DrawShadows;
    procedure Init;
    procedure SetDCPixelFormat;
    procedure OnIdleHandler(Sender: TObject; var Done: Boolean);
  protected
    //procedure WMPaint(var Msg: TWMPaint); message WM_PAINT;
  end;

const
  Earth = 1;
  Moon = 2;
  Sun = 3;

var
  frmGL: TfrmGL;
  Angle: GLfloat = 0;
  Position: GLfloat = 0.0;
  time: LongInt;
  x: extended = 0;
  mx,my: integer;
  grad: integer;

implementation


{$R *.DFM}

function PrepareImage(bmap: string):GLuint;
type
  PPixelArray = ^TPixelArray;
  TPixelArray = array [0..$7ffffff] of Byte;
var
  //Picture: TPicture;
  Picture: TJPEGImage;

  Bitmap: TBitmap;
  Data: PPixelArray;
  //Width, Height: boolean;
  i,j:  Integer;
  Temp: cardinal;
  Tex: GLuint;

begin
  Picture:= TJPEGImage.Create;
  Bitmap:= TBitmap.Create;
  Data:=nil;
  Tex:=0;
  try
    Picture.LoadFromFile(bmap);
    Bitmap.Width:= Picture.Width;
    Bitmap.Height:= Picture.Height;
    //Canvas.Draw(0,0,Picture);
    Bitmap.Canvas.Draw(0, 0, Picture);
    GetMem (Data, Bitmap.Width * Bitmap.Height * 3);
    for i:= 0 to Bitmap.Height-1 do begin
      for j:= 0 to Bitmap.Width-1 do begin
         Temp:=Bitmap.Canvas.Pixels[j,i];
         Data^[(i*Bitmap.Width+j)*3]:=Temp and 255;
         Data^[(i*Bitmap.Width+j)*3+1]:=(Temp shr 8) and 255;
         Data^[(i*Bitmap.Width+j)*3+2]:=(Temp shr 16) and 255;
       end;
    end;
    glGenTextures(1, Tex);
    glBindTexture(GL_TEXTURE_2D, Tex);
    gluBuild2DMipmaps(
      GL_TEXTURE_2D, 3,
      Bitmap.Width, Bitmap.Height,
      GL_RGB, GL_UNSIGNED_BYTE,
      Data
    );
    if 0<>glGetError then
      raise Exception.Create('Invalid image: ' + bmap);
  finally
    PrepareImage:=Tex;
    FreeMem (Data);
    Bitmap.Free;
    Picture.Free;
  end;
end;

function PrepareImageAlpha(bmap: string):GLuint;
type
  PPixelArray = ^TPixelArray;
  TPixelArray = array [0..$7ffffff] of Byte;
var
  //Picture: TPicture;
  Picture: TJPEGImage;

  Bitmap: TBitmap;
  Data: PPixelArray;
  //Width, Height: boolean;
  i,j:  Integer;
  Temp: cardinal;
  Tex: GLuint;

begin
  Picture:= TJPEGImage.Create;
  Bitmap:= TBitmap.Create;
  Data:=nil;
  Tex:=0;
  try
    Picture.LoadFromFile(bmap);
    Bitmap.Width:= Picture.Width;
    Bitmap.Height:= Picture.Height;
    //Canvas.Draw(0,0,Picture);
    Bitmap.Canvas.Draw(0, 0, Picture);
    GetMem (Data, Bitmap.Width * Bitmap.Height * 4);
    for i:= 0 to Bitmap.Height-1 do begin
      for j:= 0 to Bitmap.Width-1 do begin
         Temp:=Bitmap.Canvas.Pixels[j,i];
         Data^[(i*Bitmap.Width+j)*4]:=Temp and 255;
         Data^[(i*Bitmap.Width+j)*4+1]:=(Temp shr 8) and 255;
         Data^[(i*Bitmap.Width+j)*4+2]:=(Temp shr 16) and 255;
         Data^[(i*Bitmap.Width+j)*4+3]:=
           (integer(Data^[(i*Bitmap.Width+j)*4])+
           integer(Data^[(i*Bitmap.Width+j)*4+1])+
           integer(Data^[(i*Bitmap.Width+j)*4+2])) div 3;

       end;
    end;
    glGenTextures(1, Tex);
    glBindTexture(GL_TEXTURE_2D, Tex);
    gluBuild2DMipmaps(
      GL_TEXTURE_2D, 4,
      Bitmap.Width, Bitmap.Height,
      GL_RGBA, GL_UNSIGNED_BYTE,
      Data
    );
    if 0<>glGetError then
      raise Exception.Create('Invalid image: ' + bmap);
  finally
    PrepareImageAlpha:=Tex;
    FreeMem (Data);
    Bitmap.Free;
    Picture.Free;
  end;
end;



{ TLight }

procedure TLight.LoadToGL(LIGHTx: integer);
begin
  glLightfv(GL_LIGHT0 + LIGHTx, GL_DIFFUSE, @col.x);
  glLightfv(GL_LIGHT0 + LIGHTx, GL_POSITION, @pos.x);
  glLightf(GL_LIGHT0 + LIGHTx, GL_CONSTANT_ATTENUATION, 1);
  glLightf(GL_LIGHT0 + LIGHTx, GL_LINEAR_ATTENUATION, 0);
  glLightf(GL_LIGHT0 + LIGHTx, GL_QUADRATIC_ATTENUATION, 0);
end;


//=================== TObj ==========================================

constructor TObj.Create(frmGL: TfrmGL);
begin
  self.frmGL:=frmGL;
  SubObj:=TObjectList.Create;
end;


destructor TObj.Destroy;
begin
  SubObj.Free;
  inherited Destroy;
end;


procedure TObj.Draw;
var
  i: integer;
begin
  for i:= 0 to SubObj.Count-1 do
    (SubObj.Items[i] as TObj).Draw();
end;

{ TSphere }

constructor TSphere.Create(frmGL: TfrmGL; sr:RealType; TexName:string);
var
  Quadric: gluQuadricObj;
begin
  inherited Create(frmGL);
  r:=sr;
  if (TexName <> '') then
    Tex:=PrepareImage(TexName);
  list:=glGenLists(1);
  glNewList(list, GL_COMPILE);
    Quadric:=gluNewQuadric();
    if (TexName <> '') then
      gluQuadricTexture (Quadric, true);
    gluSphere(Quadric, sr, 20, 20);
    gluDeleteQuadric(Quadric);
  glEndList();
end;


procedure TSphere.Draw;
var
  v: vector;
begin
  glBindTexture(GL_TEXTURE_2D, Tex);
  glCallList(list);
  v.Let(0,0,0);
  pos:=mtMul(v, frmGL.CurrMatrix);
  inherited;
end;


procedure TSphere.CastShadows(const l: Vector);
const
  m=20;
var
  poly: array[0..m] of Vector;
  poly2: array[0..m] of Vector;
  n, t, t2, a, b, c, x: Vector;
  sina, cosa, sinug, cosug, va, vb, vc: RealType;
  i: integer;
begin
  n:=mtSub(l, pos);
  a:=n; a.Normalize;
  glLoadIdentity;
  if sqr(n.x)<sqr(n.y) then begin
    t.Let(1,0,0);
  end else begin
    t.Let(0,1,0);
  end;
  b:=cMul(t, n);
  b.Normalize;
  sina:= r/ n.Length;
  cosa:= sqrt(1-sqr(sina));
  {x:=rMul(
    b,
    2*sina/sqrt(1-sqr(sina))*n.Length
  );}
  c:=cMul(a,b);
  va:=r*sina;
  t:=mtAdd(x, pos);
  glBegin(GL_QUAD_STRIP);
  for i:=0 to m do begin
    sinug:=sin(i/m*2*pi);
    cosug:=cos(i/m*2*pi);
    vb:=sinug*r*cosa;
    vc:=cosug*r*cosa;
    t:=mtAdd(
      rMul(a, va), mtAdd(
      rMul(b, vb),
      rMul(c, vc))
    );
    t:=mtAdd(pos, t);
    t2:=t;
    t2:=mtSub(t2, l);
    t2:=mtAdd(t2, t);
    poly[i]:=t2;
    poly2[i]:=t;

    glColor3f(0,1,0);
      glVertex3fv(@t2.x);
      glVertex3fv(@t.x);
    //pos + cosug*r*cosa*b + sinug*r*cosa*c + va*a
  end;
  glEnd;

  glBegin(GL_POLYGON);
    for i:=0 to m-1 do begin
      glVertex3fv(@poly[i].x);
    end;
  glEnd;
  glBegin(GL_POLYGON);
    for i:=0 to m-1 do begin
      glVertex3fv(@poly2[m-i-1].x);
    end;
  glEnd;
end;


{ TPseudoEarth }

constructor TPseudoEarth.Create(frmGL: TfrmGL);
begin
  inherited;
  Orbit:=30;
  Year:=365.25;
end;

procedure TPseudoEarth.Draw;
var
  ugol: RealType;
  m:Matrix;
begin
  frmGL.Save(m);
  ugol:= -frmGL.Time/Year*2*pi;
  frmGL.Mul(mtTranslate(
    Orbit*cos(ugol), Orbit*sin(ugol), 0
  ));
  inherited;
  frmGL.Restore(m);
end;


{ TAtmosphere }
procedure TAtmosphere.Draw;
const
  diffuse: array [0..3] of RealType=(0, 0.8, 1, 0.5);
  diffuse1: array [0..3] of RealType=(1, 1, 1, 1);
var
  m:Matrix;
begin
  frmGL.Save(m);
  glDepthMask(GL_FALSE);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, @diffuse[0]);
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE);
  inherited;
  glDisable(GL_BLEND);
  glDepthMask(GL_TRUE);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, @diffuse1[0]);
  frmGL.Restore(m);
end;

{ TEarth }
procedure TEarth.Draw;
var
  m:Matrix;
begin
  frmGL.Save(m);
  frmGL.Mul(mtMul(mtRotY(-22.5/180*pi),mtRotZ(-frmGL.Time*2*pi)));
  //frmGL.Mul(mtRotZ(-frmGL.Time));
  inherited;
  frmGL.Restore(m);
end;

{ TMoon }

constructor TMoon.Create(frmGL: TfrmGL; sr: RealType; TexName: string);
begin
  inherited;
  Orbit:=12;
  Year:=29;
end;

procedure TMoon.Draw;
var
  m:Matrix;
  ugol: RealType;
begin
  frmGL.Save(m);
  ugol:= -frmGL.Time/Year*2*pi;
  frmGL.Mul(mtMul(mtRotZ(ugol), mtTranslate(Orbit, 0, 0)));
  inherited;
  frmGL.Restore(m);
end;

{ Particle }

procedure Particle.Born;
begin
  size:=4;
  max:=random(1000)/300;
  live:=max;
  pos.Let(
    random(100)-50,
    random(100)-50,
    random(100)-50
  );
  pos.Normalize;
  pos:=rMul(pos, random(100)/100+3);

  v:=pos;
  v.Normalize;
  v:=rMul(v, random(100)/100);
end;

procedure Particle.Next(dt: RealType);
begin
  live:=live-dt;
  if (live < 0) then begin
    Born;
    exit;
  end;
  pos.x:=pos.x+v.x*dt;
  pos.y:=pos.y+v.y*dt;
  pos.z:=pos.z+v.z*dt;
end;

procedure Particle.Draw(var m: Matrix);
var
  centr: Vector;
  size_2: RealType;
  t: RealType;
begin
  //MODELVIEW in gl should be identity
  //m should be real MODELVIEW.
  //OpenGL should be in glBegin(GL_QUADS) mode
  t:=Live/Max;
  centr:= mtMul(pos, m);
  size_2:=size*0.5*t;

  centr.x:=centr.x+size_2;
  centr.y:=centr.y+size_2;
  glTexCoord2f(1,1);
  glVertex3fv(@centr.x);
  centr.x:=centr.x-size*t;
  glTexCoord2f(0,1);
  glVertex3fv(@centr.x);
  centr.y:=centr.y-size*t;
  glTexCoord2f(0,0);
  glVertex3fv(@centr.x);
  centr.x:=centr.x+size*t;
  glTexCoord2f(1,0);
  glVertex3fv(@centr.x);
end;


{ TSun }

constructor TSun.Create(frmGL: TfrmGL);
var
  i: integer;
begin
  inherited;
  n:=400;
  SetLength(ps, n);
  for i:=0 to n-1 do begin
    ps[i].Born;
  end;
  tex_id:=PrepareImageAlpha('../sun.jpg');
end;

destructor TSun.Destroy;
begin
  ps:=nil;
  inherited;
end;


procedure TSun.Draw;
var
  m: Matrix;
  i: integer;
begin
  frmGL.Save(m);
  frmGL.Mul(mtRotZ(-frmGL.Time/40));
  glDisable(GL_LIGHTING);
  //glDisable(GL_DEPTH_TEST);
  //glDepthFunc(GL_ALWAYS);
  glDepthMask(GL_FALSE);

  glEnable(GL_BLEND);
  glEnable(GL_ALPHA_TEST);
  //glAlphaFunc(GL_GREATER, 0.1);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE);
  glLoadIdentity;
  glBindTexture(GL_TEXTURE_2D, tex_id);
  glColor3f(1,1,1);
  glBegin(GL_QUADS);
    for i:=0 to n-1 do begin
      ps[i].Draw(frmGL.CurrMatrix);
    end;
  glEnd;
  for i:=0 to n-1 do begin
    ps[i].Next(frmGL.Time-frmGL.LastTime);
  end;
  inherited;
  glDisable(GL_ALPHA_TEST);
  glDisable(GL_BLEND);
  //glEnable(GL_DEPTH_TEST);
  //glDepthFunc(GL_LESS);
  glDepthMask(GL_TRUE);
  glEnable(GL_LIGHTING);
  frmGL.Restore(m);
end;

procedure TSky.Draw;
var
  m:Matrix;
begin
  frmGL.Save(m);
  frmGL.CurrMatrix.m[3,0]:=0;
  frmGL.CurrMatrix.m[3,1]:=0;
  frmGL.CurrMatrix.m[3,2]:=0;
  mtLoadToGL(frmGL.CurrMatrix);
  glDisable(GL_LIGHTING);
  glDisable(GL_DEPTH_TEST);
  glMatrixMode(GL_TEXTURE);
  glScalef(0.99, 0.99, 1);
  glColor3f(1,1,1);
  glCullFace(GL_FRONT);
  inherited;
  glCullFace(GL_BACK);
  glLoadIdentity;
  glMatrixMode(GL_MODELVIEW);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_LIGHTING);
  frmGL.Restore(m);
end;



{======================================================================
 }



{=======================================================================
}
procedure TfrmGL.Init;
{const
 LightPos : Array [0..3] of GLFloat = (10.0, 10.0, 10.0, 1.0);
 SunColor : Array [0..3] of GLFloat = (11.0, 2.9, 0.0, 0.0);
var
 Quadric : GLUquadricObj;}
begin
 {glEnable(GL_LIGHTING);
 glEnable(GL_LIGHT0);
 glLightfv(GL_LIGHT0, GL_POSITION, @LightPos);

 Quadric := gluNewQuadric;
 gluQuadricTexture (Quadric, TRUE);

 glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
 glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

 glEnable(GL_TEXTURE_2D);


 //
 glNewList (Sun, GL_COMPILE);
   glEnable(GL_DEPTH_TEST);
   glLightModelfv(GL_LIGHT_MODEL_AMBIENT, @SunColor);
   glPushMatrix;
   gluSphere (Quadric, 5, 24, 24);
   glPopMatrix;
   //glEnable(GL_LIGHT1);
 glEndList;

 //
 glNewList (Earth, GL_COMPILE);
   prepareImage ('..\earth.jpg');
   glPushMatrix;
   gluSphere (Quadric, 3, 24, 24);
   glPopMatrix;
 glEndList;
 //
 glNewList (Moon, GL_COMPILE);
   prepareImage ('..\moon.jpg');
   glPushMatrix;
   //glTranslatef(0,cos(Angle/180.0*3.14)*2.5, sin(Angle/180.0*3.14)*2.0);
   glColor3f(0.5, 0.5, 1.0);
   gluSphere (Quadric, 1, 24, 24);
   glPopMatrix;
 glEndList;

 gluDeleteQuadric (Quadric);

 glEnable(GL_DEPTH_TEST);}
end;

{=======================================================================
 }
{procedure povorot;
begin
     glRotatef (-Angle*40, 0.0, 1.0, 0.0);
     glTranslatef(50,0,0);
     glPushMatrix;
     glRotate(-Angle*200, 0, 1, 0);
     glRotate(22.5, 1, 0, 0);
     glCallList(Earth);
     glPopMatrix;

     glRotatef(-Angle*100, 0.0, 1.0, 0.0);
     glTranslatef(20,2.2,0.2);
     //glRotatef (-10, 0.0, 1.0, 0.0);
     glCallList(Moon);
end;}

{procedure TfrmGL.WMPaint(var Msg: TWMPaint);
var
  ps : TPaintStruct;

begin
  BeginPaint(Handle, ps);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity;
  Angle:= (GetTickCount-StartTime)/1000.0;

  glClear( GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT );

  glTranslate(0,0,-120);
  glCallList(Sun);

  povorot;

  SwapBuffers(DC);
  EndPaint(Handle, ps);

  //Position := Position;// + 0.25 * 360 / 1000;
  InvalidateRect(Handle, nil, False);
end;
}
{=======================================================================
 }
procedure TfrmGL.FormCreate(Sender: TObject);
begin
  Application.OnIdle:=OnIdleHandler;
  DC := GetDC(Handle);
  SetDCPixelFormat;
  hrc := wglCreateContext(DC);
  wglMakeCurrent(DC, hrc);
  Init;
  StartTime := GetTickCount;
  mx:=0;
  my:=0;
  grad:=5;

  TimeToAwake:=12;

  Distance:=60;

  MouseMatrix:=mtIdentity;

  Light:=TLight.Create();
  Light.pos.Let(0,0,0);
  Light.col.Let(1,1,0.8);

  Scene:= TObj.Create(self);
    Sky:= TSky.Create(self, 10, '../sky2.jpg');
    Scene.SubObj.Add(Sky);

    PseudoEarth:=TPseudoEarth.Create(self);
    Scene.SubObj.Add(PseudoEarth);
      Earth:= TEarth.Create(self, 4, '../earth.jpg');
      PseudoEarth.SubObj.Add(Earth);

      Moon:= TMoon.Create(self, 2, '../moon.jpg');
      PseudoEarth.SubObj.Add(Moon);

      Atmosphere:=TAtmosphere.Create(self, Earth.r*1.05, '');
      PseudoEarth.SubObj.Add(Atmosphere);

  Sun:= TSun.Create(self);
    //Scene.SubObj.Add(Sun);

  ShadowCasters:=TObjectList.Create(false);
  ShadowCasters.Add(Moon);
  ShadowCasters.Add(Earth);
end;

procedure TfrmGL.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  Scene.Free;
  Light.Free;
  Sun.Free;
  ShadowCasters.Free;
  wglMakeCurrent(0, 0);
  wglDeleteContext(hrc);
  ReleaseDC(Handle, DC);
  DeleteDC (DC);
end;

{=======================================================================
  }
procedure TfrmGL.FormResize(Sender: TObject);
begin
 glViewport(0, 0, Width, Height );
 glMatrixMode( GL_PROJECTION );
 glLoadIdentity;
 //glFrustum( -3, 3, -3, 3, 7, 1500.0 );
 gluPerspective(60, Width/(Height+1), 1, 200);
 InvalidateRect(Handle, nil, False);
end;

{=======================================================================
  }


{=======================================================================
  }
procedure TfrmGL.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  If Key = VK_ESCAPE then Close;
end;

{=======================================================================
  }
procedure TfrmGL.SetDCPixelFormat;
var
  nPixelFormat: Integer;
  pfd: TPixelFormatDescriptor;

begin
  FillChar(pfd, SizeOf(pfd), 0);

  pfd.nSize:=sizeof(pfd);
  pfd.dwFlags := PFD_DRAW_TO_WINDOW or PFD_SUPPORT_OPENGL or
                 PFD_DOUBLEBUFFER;
  pfd.nVersion:=1;
  pfd.dwFlags:=
    PFD_DRAW_TO_WINDOW or
    PFD_SUPPORT_OPENGL or
    PFD_DOUBLEBUFFER ;
  pfd.iPixelType:=PFD_TYPE_RGBA;
  pfd.cColorBits:=24;
  pfd.cDepthBits:=32;
  pfd.cStencilBits:=8;  
  pfd.iLayerType:=PFD_MAIN_PLANE;


  nPixelFormat := ChoosePixelFormat(DC, @pfd);
  SetPixelFormat(DC, nPixelFormat, @pfd);
end;


procedure TfrmGL.FormMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  DC:HDC;
  i,j: integer;
begin
  if Button=mbLeft then
    LeftMouse:= true;
  if Button=mbRight then
    RightMouse:= true;
  OldMouseY:= X;
  OldMouseY:= Y;
end;

procedure TfrmGL.FormMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if Button=mbLeft then
    LeftMouse:= false;
  if Button=mbRight then
    RightMouse:= false;
end;

procedure TfrmGL.FormMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
var
  MouseDy, MouseDx: integer;
begin
  MouseDx:= X-OldMouseX;
  MouseDy:= Y-OldMouseY;
  if LeftMouse then
  begin
    MouseMatrix:=
      mtMul(
        mtMul(
          mtRotX(MouseDy/100),
          mtRotY(MouseDx/100)
        ),
        MouseMatrix
      );
  end;
  if RightMouse then
    Distance:= Distance + MouseDy;
  OldMouseX:=X;
  OldMouseY:=Y;
end;


procedure TfrmGL.OnIdleHandler(Sender: TObject; var Done: Boolean);
var
  m: Matrix;
  msg: array[0..14] of char;
  i:integer;
begin
  glMatrixMode(GL_MODELVIEW);
  CurrMatrix:=mtIdentity;
  mtLoadToGL(CurrMatrix);
  Mul(mtTranslate(0, 0, -Distance));
  Mul(MouseMatrix);
  glClear( GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT );
  LastTime:=Time;
  Time:=(GetTickCount-StartTime)/1000;
  glEnable(GL_DEPTH_TEST);

  glEnable(GL_CULL_FACE);
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  Light.LoadToGL(0);
  LightPos:=mtMul(Light.pos, CurrMatrix);
  glEnable(GL_TEXTURE_2D);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);

    Scene.Draw();
  glDisable(GL_TEXTURE_2D);
  glDisable(GL_LIGHT0);
  glDisable(GL_LIGHTING);

  Save(m);
  glEnable(GL_STENCIL_TEST);
    GenShadows;
    DrawShadows;
  glDisable(GL_STENCIL_TEST);
  glMatrixMode(GL_MODELVIEW);
  Restore(m);

  glEnable(GL_TEXTURE_2D);
  Sun.Draw;
  glDisable(GL_TEXTURE_2D);

  glDisable(GL_DEPTH_TEST);
  glDisable(GL_CULL_FACE);

  SwapBuffers(DC);

  Done:=false;
end;

procedure TfrmGL.GenShadows;
var
  i:integer;
begin
  glDepthMask(GL_FALSE);
  glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
  glCullFace(GL_FRONT);
  glStencilFunc(GL_ALWAYS, 0, 0);
  glStencilOp(
    GL_KEEP,
    GL_INCR,
    GL_KEEP
  );
  for i:=0 to ShadowCasters.Count-1 do
    (ShadowCasters.Items[i] as TShCaster).CastShadows(LightPos);
  glDepthMask(GL_FALSE);
  glCullFace(GL_BACK);
  glStencilOp(
    GL_KEEP,
    GL_DECR,
    GL_KEEP
  );
  for i:=0 to ShadowCasters.Count-1 do
    (ShadowCasters.Items[i] as TShCaster).CastShadows(LightPos);

end;


procedure TfrmGL.DrawShadows;
begin
  glDepthFunc(GL_ALWAYS);
  glStencilFunc(GL_NOTEQUAL, 0, 255);
  glStencilOp(
    GL_KEEP,
    GL_KEEP,
    GL_ZERO
  );
  glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity;
  glMatrixMode(GL_PROJECTION);
  glPushMatrix;
  glLoadIdentity;
    glColor3f(0, 0, 0);
    glBegin(GL_QUADS);
      glVertex2f(1,1);
      glVertex2f(-1,1);
      glVertex2f(-1,-1);
      glVertex2f(1,-1);
    glEnd;
  glPopMatrix;
  glDepthFunc(GL_LESS);
  glDepthMask(GL_TRUE);
end;


procedure TfrmGL.Mul(const m:Matrix);
begin
  CurrMatrix:=mtMul(CurrMatrix, m);
  mtMulToGl(m);
end;


procedure TfrmGL.Restore(var RestoreFrom:Matrix);
begin
  dec(sp);
  if sp<0 then
    raise self;
  CurrMatrix:=RestoreFrom;
  glPopMatrix;
end;


procedure TfrmGL.Save(var SaveTo:Matrix);
begin
  inc(sp);
  if sp>100 then
    raise self;
  SaveTo:=CurrMatrix;
  glPushMatrix;
end;



end.


